"""
Created on 09 Nov 2020

@author: Jade Page (jade.page@southcoastscience.com)
"""

from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class EmailList(PersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "device_email_list"

    @classmethod
    def persistence_location(cls):
        return cls.aws_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        email_list = jdict.get('email_list')

        return cls(email_list)


    @staticmethod
    def __item_contains_address(addresses, email_address):
        if addresses is None:
            return False

        return email_address in addresses


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, email_list):
        """
        Constructor
        """
        super().__init__()

        self.__email_list = email_list


    # ----------------------------------------------------------------------------------------------------------------

    def append(self, device_tag, email_address):
        if device_tag not in self.__email_list:
            return False

        if self.__email_list[device_tag] is None:
            self.__email_list[device_tag] = email_address
            return True

        if device_tag in self.__email_list[device_tag]:
            return True

        try:
            self.__email_list[device_tag].append(email_address)
            self.__email_list[device_tag].sort()

        except AttributeError:
            self.__email_list[device_tag] = sorted([email_address, self.__email_list[device_tag]])

        return True


    def remove(self, device_tag, email_address):
        # all devices...
        if device_tag is None:
            for listed_tag in self.__email_list:
                self.__remove_for_device(listed_tag, email_address)

            return True

        # specific device...
        if device_tag not in self.__email_list:
            return False

        self.__remove_for_device(device_tag, email_address)

        return True


    def __remove_for_device(self, device_tag, email_address):
        if self.__email_list[device_tag] is None:
            return

        try:
            try:
                self.__email_list[device_tag].remove(email_address)
            except ValueError:
                pass

        except AttributeError:
            if self.__email_list[device_tag] == email_address:
                self.__email_list[device_tag] = None


    # ----------------------------------------------------------------------------------------------------------------

    def subset(self, email_address):
        email_list = {item: addresses for item, addresses in self.__email_list.items()
                      if self.__item_contains_address(addresses, email_address)}

        return EmailList(email_list)


    def emails(self, device_tag):
        try:
            return self.__email_list[device_tag]
        except KeyError:
            return None


    @property
    def email_list(self):
        return self.__email_list


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        return {'email_list': self.__email_list}


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "EmailList:{email_list:%s}" %  self.__email_list



"""
Created on 2 Jul 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
    {
      "client_id": "5926",
      "owner": "southcoastscience-dev",
      "topic": "/orgs/south-coast-science-dev/development/loc/3/gases",
      "date": "2017-03-19T08:00:18.414Z",
      "payload": {
        "encoding": "utf-8",
        "content-type": "application/json",
        "text": "{"tag": "scs-ap1-6", "rec": "2017-03-19T08:00:17.601+00:00", "val":
        {"NO2": {"weV": 0.286379, "aeV": 0.288317, "weC": -0.014025, "cnc": -76.2},
        "Ox": {"weV": 0.416506, "aeV": 0.407381, "weC": null, "cnc": null},
        "NO": {"weV": 0.296255, "aeV": 0.31138, "weC": -0.036358, "cnc": -98.5},
        "CO": {"weV": 0.33413, "aeV": 0.566696, "weC": 0.34805, "cnc": 1456.3},
        "pt1": {"v": 0.324005, "tmp": 23.3},
        "sht": {"hmd": 45.4, "tmp": 21.9}}}"
      },
"""

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class DeviceTopic(JSONable):
    """
    classdocs
   """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_message_jdict(cls, jdict):
        if not jdict:
            return None

        client_id = jdict.get('device')
        path = jdict.get('topic')
        earliest_publication = LocalizedDatetime.construct_from_jdict(jdict.get('date'))

        client_id = DeviceTopic(client_id, path, earliest_publication)

        return client_id


    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        client_id = jdict.get('client-id')
        path = jdict.get('path')
        earliest_publication = LocalizedDatetime.construct_from_jdict(jdict.get('earliest-pub'))

        client_id = DeviceTopic(client_id, path, earliest_publication)

        return client_id


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, client_id, path, earliest_publication):
        """
        Constructor
        """
        self.__client_id = client_id                            # string (int by convention)
        self.__path = path                                      # string
        self.__earliest_publication = earliest_publication      # LocalisedDatetime


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['client-id'] = self.client_id
        jdict['path'] = self.path
        jdict['earliest-pub'] = self.earliest_publication

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def client_id(self):
        return self.__client_id


    @property
    def path(self):
        return self.__path


    @property
    def earliest_publication(self):
        return self.__earliest_publication


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "DeviceTopic:{client_id:%s, path:%s, earliest_publication:%s}" % \
               (self.client_id, self.path, self.earliest_publication)



"""
Created on 14 Oct 2021

@author: Jade Page (jade.page@southcoastscience.com)

A class holding RDS functionality

https://johnkeefe.net/amazon-aurora-mysql-plus-python

https://stackoverflow.com/questions/59742083/passing-in-iam-credentials-when-using-the-aurora-serverless-data-api

https://github.com/cal-poly-dxhub/covid-dashboard/blob/1d4d3f9abe82e84c22d294642e733b5bfffb2780/lambda/utility.py

"""
# --------------------------------------------------------------------------------------------------------------------
import ast
import json
import logging

# --------------------------------------------------------------------------------------------------------------------



class RDSManager(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, rds_client, secret_arn, database, resource_arn):
        self.__rds_client = rds_client  # session.client('rds')
        self.__secretArn = secret_arn
        self.__database = database
        self.__resourceArn = resource_arn

    # ----------------------------------------------------------------------------------------------------------------

    def execute_query_raw(self, q):
        res = self.__rds_client.execute_statement(
            secretArn=self.__secretArn,
            database=self.__database,
            resourceArn=self.__resourceArn,
            sql=q
        )
        logging.info(res)
        return res

    def execute_query_mapped(self, q):
        res = self.__rds_client.execute_statement(
            includeResultMetadata=True,
            secretArn=self.__secretArn,
            database=self.__database,
            resourceArn=self.__resourceArn,
            sql=q
        )
        return self.generate_map_from_response(res)

    @staticmethod
    def generate_map_from_response(response):
        response_set = []
        obj = {}
        for record in response['records']:
            for i in range(len(record)):
                obj[response['columnMetadata'][i]['label']] = list(record[i].values())[0]
            response_set.append(obj.copy())
        return response_set

# --------------------------------------------------------------------------------------------------------------------


class RDSLambdaManager(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, lambda_client):
        self.__lambda_client = lambda_client


    def get_raw_response(self, q, user):
        event = {
            "query": q,
            "username": user
        }
        payload = json.dumps(event).encode('utf-8')

        res = self.__lambda_client.invoke(
            FunctionName="arn:aws:lambda:us-west-2:696437392763:function:RDSExecutor",
            InvocationType='RequestResponse',
            Payload=payload
        )

        return res

    def get_records(self, q, user):
        event = {
            "query": q,
            "username": user
        }
        payload = json.dumps(event).encode('utf-8')

        res = self.__lambda_client.invoke(
            FunctionName="arn:aws:lambda:us-west-2:696437392763:function:RDSExecutor",
            InvocationType='RequestResponse',
            Payload=payload
        )

        raw = res['Payload'].read().decode('utf-8')
        jdict = json.loads(ast.literal_eval(raw))
        logging.info(jdict)
        if type(jdict) is str:
            return None
        return jdict

    def do_query(self, q, user):
        event = {
            "query": q,
            "username": user
        }
        payload = json.dumps(event).encode('utf-8')

        self.__lambda_client.invoke(
            FunctionName="arn:aws:lambda:us-west-2:696437392763:function:RDSExecutor",
            InvocationType='RequestResponse',
            Payload=payload
        )


    """
    TODO: Could this be made more efficient, maybe by getting one row with raw metadata to do the 
    mapping and the rest without? Is there any guarantee all calls are extracted in the same order???
    """

    @staticmethod
    def do_mapping(response):
        response_set = []
        obj = {}
        for record in response['records']:
            for i in range(len(record)):
                obj[response['columnMetadata'][i]['label']] = list(record[i].values())[0]
            response_set.append(obj.copy())
        return response_set



"""
Created on 10 Nov 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
{
    "period": 604800,
    "last-reading": "5 seconds ago",
    "average-frequency": "9 seconds",
    "total": 49661,
    "contributors": [
      {
        "name": "South Coast Science - Dev",
        "id": "southcoastscience-dev",
        "gravatar-hash": "07f512e9fe64863039df0c0f1834cc25"
      }
    ]
    "last-location": {
      "lat": 50.819456,
      "lon": -0.128336
    }
}
"""

from collections import OrderedDict

from scs_core.data.json import JSONable
from scs_core.data.str import Str

from scs_core.osio.data.topic_contributor import TopicContributor
from scs_core.osio.data.location import Location


# --------------------------------------------------------------------------------------------------------------------

class TopicStats(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        period = jdict.get('period')
        last_reading = jdict.get('last-reading')
        average_frequency = jdict.get('average-frequency')
        total = jdict.get('total')

        contributors_jdict = jdict.get('contributors')

        contributors = [TopicContributor.construct_from_jdict(contributor_jdict)
                        for contributor_jdict in contributors_jdict] if contributors_jdict else []

        last_location = Location.construct_from_jdict(jdict.get('last-location'))

        return TopicStats(period, last_reading, average_frequency, total, contributors, last_location)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, period, last_reading, average_frequency, total, contributors, last_location):
        """
        Constructor
        """
        self.__period = period                          # int
        self.__last_reading = last_reading              # string
        self.__average_frequency = average_frequency    # string
        self.__total = total                            # int

        self.__contributors = contributors              # list of TopicContributor
        self.__last_location = last_location            # Location


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['period'] = self.period
        jdict['last-reading'] = self.last_reading
        jdict['average-frequency'] = self.average_frequency
        jdict['total'] = self.total

        jdict['contributors'] = self.contributors
        jdict['last-location'] = self.last_location

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def period(self):
        return self.__period


    @property
    def last_reading(self):
        return self.__last_reading


    @property
    def average_frequency(self):
        return self.__average_frequency


    @property
    def total(self):
        return self.__total


    @property
    def contributors(self):
        return self.__contributors


    @property
    def last_location(self):
        return self.__last_location


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "TopicStats:{period:%s, last_reading:%s, average_frequency:%s, total:%s, " \
               "contributors:%s, last_location:%s}" % \
               (self.period, self.last_reading, self.average_frequency, self.total,
                Str.collection(self.contributors), self.last_location)



"""
Created on 27 Jan 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document:
{
    "rec": "2021-11-02T10:53:15Z",
    "tag": "scs-opc-1",
    "ver": 1.0,
    "val": {
        "hostname": "scs-cube-001",
        "packs": {
            "scs_core": {
                "repo": "scs_core",
                "version": "1.1.1"
            },
            "scs_dev": {
                "repo": "scs_dev",
                "version": "1.1.1"
            },
            "scs_dfe": {
                "repo": "scs_dfe_eng",
                "version": "1.1.0"
            },
            "scs_greengrass": {
                "repo": "scs_greengrass",
                "version": "2.1.2"
            },
            "scs_host": {
                "repo": "scs_host_cpc",
                "version": "1.0.5"
            },
            "scs_inference": {
                "repo": "scs_inference",
                "version": null
            },
            "scs_mfr": {
                "repo": "scs_mfr",
                "version": "1.1.1"
            },
            "scs_ndir": {
                "repo": "scs_ndir",
                "version": null
            },
            "scs_psu": {
                "repo": "scs_psu",
                "version": "1.0.10"
            }
        },
        "afe-baseline": {
            "sn1": {
                "calibrated-on": "2021-11-01T17:43:41Z",
                "offset": 120
            },
            "sn2": {
                "calibrated-on": "2021-11-01T17:32:48Z",
                "offset": 0
            },
            "sn3": {
                "calibrated-on": "2021-11-01T17:32:48Z",
                "offset": 0
            },
            "sn4": {
                "calibrated-on": "2021-11-01T17:32:48Z",
                "offset": 0
            }
        },
        "afe-id": {
            "serial_number": null,
            "type": "DSI",
            "calibrated_on": "2020-01-01",
            "sn1": {
                "serial_number": "212060325",
                "sensor_type": "NOGA4"
            }
        },
        "aws-api-auth": {
            "endpoint": "aws.southcoastscience.com",
            "api-key": "de92c5ff-b47a-4cc4-a04c-62d684d64a1f"
        },
        "aws-group-config": {
            "group-name": "scs-cube-001-group",
            "time-initiated": "2021-02-08T16:38:55Z",
            "unix-group": 984,
            "ml": true
        },
        "aws-project": {
            "location-path": "south-coast-science-dev/cube/loc/1",
            "device-path": "south-coast-science-dev/cube/device"
        },
        "csv-logger-conf": {
            "root-path": "/srv/removable_data_storage",
            "delete-oldest": true,
            "write-interval": 0
        },
        "display-conf": null,
        "vcal-baseline": {
            "NO2": {
                "calibrated-on": "2021-11-01T17:33:10Z",
                "offset": 107
            }
        },
        "gas-baseline": {
            "NO2": {
                "calibrated-on": "2021-11-01T09:30:50Z",
                "offset": 0
            }
        },
        "gas-model-conf": {
            "uds-path": "pipes/lambda-gas-model.uds",
            "model-interface": "vE",
            "model-compendium-group": "OE21HA"
        },
        "gps-conf": {
            "model": "SAM8Q",
            "sample-interval": 10,
            "tally": 60,
            "report-file": "/tmp/southcoastscience/gps_report.json",
            "debug": false
        },
        "interface-conf": {
            "model": "OPCubeT1"
        },
        "greengrass-identity": {
            "core-name": "scs-cube-001-core",
            "group-name": "scs-cube-001-group"
        },
        "mpl115a2-calib": null,
        "mqtt-conf": {
            "inhibit-publishing": false,
            "report-file": "/tmp/southcoastscience/mqtt_queue_report.json",
            "debug": false
        },
        "ndir-conf": null,
        "opc-conf": {
            "model": "N3",
            "sample-period": 10,
            "restart-on-zeroes": true,
            "power-saving": false
        },
        "opc-version": {
            "serial": "177336702",
            "firmware": "OPC-N3 Iss1.1 FirmwareVer=1.17a...........................BS"
        },
        "pmx-model-conf": {
            "uds-path": "pipes/lambda-pmx-model.uds",
            "model-interface": "s2"
        },
        "pressure-conf": null,
        "psu-conf": {
            "model": "OPCubeV1",
            "batt-model": "PackV2",
            "ignore-threshold": true,
            "reporting-interval": 5,
            "report-file": "/tmp/southcoastscience/psu_status_report.json"
        },
        "psu-version": {
            "id": "SCS OPCube Controller type 1 firmware 1",
            "tag": "001.001.003",
            "c-date": null,
            "c-time": null
        },
        "pt1000-calib": null,
        "scd30-baseline": null,
        "scd30-conf": null,
        "schedule": {
            "scs-climate": {
                "interval": 60.0,
                "tally": 1
            },
            "scs-gases": {
                "interval": 10.0,
                "tally": 1
            },
            "scs-particulates": {
                "interval": 10.0,
                "tally": 1
            },
            "scs-status": {
                "interval": 60.0,
                "tally": 1
            }
        },
        "shared-secret": {
            "key": "Ytzglk6oYpzJY0FB"
        },
        "sht-conf": {
            "int": "0x45",
            "ext": "0x45"
        },
        "networks": {
            "eth0": {
                "kind": "ethernet",
                "state": "connected",
                "connection": "Ethernet eth0"
            },
            "cdc-wdm0": {
                "kind": "gsm",
                "state": "connecting",
                "connection": "(prepare)  Claro_Brazil"
            }
        },
        "modem": {
            "id": "3f07553c31ce11715037ac16c24ceddcfb6f7a0b",
            "imei": "867962041294151",
            "mfr": "QUALCOMM INCORPORATED",
            "model": "QUECTEL Mobile Broadband Module",
            "rev": "EC21EFAR06A01M4G"
        },
        "sim": {
            "imsi": "234104886708667",
            "iccid": "8944110068256270054",
            "operator-code": null,
            "operator-name": null
        },
        "system-id": {
            "set-on": "2020-09-17T11:16:25Z",
            "vendor-id": "SCS",
            "model-id": "OPC",
            "model": "Praxis/OPCube",
            "config": "v1",
            "system-sn": 1
        },
        "timezone-conf": {
            "set-on": "2017-08-15T12:50:05Z",
            "name": "Europe/London"
        }
    }
}
"""

import json
import socket

from collections import OrderedDict

from scs_core.aws.client.api_auth import APIAuth
from scs_core.aws.config.project import Project
from scs_core.aws.greengrass.aws_group_configuration import AWSGroupConfiguration
from scs_core.aws.greengrass.aws_identity import AWSIdentity

from scs_core.climate.mpl115a2_calib import MPL115A2Calib
from scs_core.climate.pressure_conf import PressureConf
from scs_core.climate.sht_conf import SHTConf

from scs_core.comms.mqtt_conf import MQTTConf

from scs_core.csv.csv_logger_conf import CSVLoggerConf

from scs_core.data.json import JSONable

from scs_core.display.display_conf import DisplayConf

from scs_core.estate.package_version import PackageVersions

from scs_core.gas.afe_baseline import AFEBaseline
from scs_core.gas.afe_id import AFEId
from scs_core.gas.afe.pt1000_calib import Pt1000Calib
from scs_core.gas.ndir.ndir_conf import NDIRConf
from scs_core.gas.scd30.scd30_conf import SCD30Conf
from scs_core.gas.scd30.scd30_baseline import SCD30Baseline

from scs_core.gps.gps_conf import GPSConf

from scs_core.interface.interface_conf import InterfaceConf

from scs_core.location.timezone_conf import TimezoneConf

from scs_core.model.gas.gas_baseline import GasBaseline
from scs_core.model.gas.gas_model_conf import GasModelConf
from scs_core.model.gas.vcal_baseline import VCalBaseline

from scs_core.model.pmx.pmx_model_conf import PMxModelConf

from scs_core.particulate.opc_conf import OPCConf
from scs_core.particulate.opc_version import OPCVersion

from scs_core.psu.psu_conf import PSUConf
from scs_core.psu.psu_version import PSUVersion

from scs_core.sync.schedule import Schedule

from scs_core.sys.modem import Modem, SIM
from scs_core.sys.network import Networks
from scs_core.sys.shared_secret import SharedSecret
from scs_core.sys.system_id import SystemID


# --------------------------------------------------------------------------------------------------------------------

class Configuration(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jstr(cls, jstr):
        try:
            return cls.construct_from_jdict(json.loads(jstr))
        except json.decoder.JSONDecodeError:
            return None


    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            if skeleton:
                return cls(None, None, None, None, None,
                           None, None, None, None, None,
                           None, None, None, None, None,
                           None, None, None, None, None,
                           None, None, None, None, None,
                           None, None, None, None, None,
                           None, None, None, None, None)
            else:
                return None

        hostname = jdict.get('hostname')
        packs = PackageVersions.construct_from_jdict(jdict.get('packs'))

        afe_baseline = AFEBaseline.construct_from_jdict(jdict.get('afe-baseline'))
        afe_id = AFEId.construct_from_jdict(jdict.get('afe-id'))
        aws_api_auth = APIAuth.construct_from_jdict(jdict.get('aws-api-auth'))
        aws_group_config = AWSGroupConfiguration.construct_from_jdict(jdict.get('aws-group-config'))
        aws_project = Project.construct_from_jdict(jdict.get('aws-project'))
        csv_logger_conf = CSVLoggerConf.construct_from_jdict(jdict.get('csv-logger-conf'))
        display_conf = DisplayConf.construct_from_jdict(jdict.get('display-conf'))
        vcal_baseline = VCalBaseline.construct_from_jdict(jdict.get('vcal-baseline'))
        gas_baseline = GasBaseline.construct_from_jdict(jdict.get('gas-baseline'))
        gas_model_conf = GasModelConf.construct_from_jdict(jdict.get('gas-model-conf'))
        gps_conf = GPSConf.construct_from_jdict(jdict.get('gps-conf'))
        greengrass_identity = AWSIdentity.construct_from_jdict(jdict.get('greengrass-identity'))
        interface_conf = InterfaceConf.construct_from_jdict(jdict.get('interface-conf'))
        mpl115a2_calib = MPL115A2Calib.construct_from_jdict(jdict.get('mpl115a2-calib'))
        mqtt_conf = MQTTConf.construct_from_jdict(jdict.get('mqtt-conf'))
        ndir_conf = NDIRConf.construct_from_jdict(jdict.get('ndir-conf'))
        opc_conf = OPCConf.construct_from_jdict(jdict.get('opc-conf'))
        opc_version = OPCVersion.construct_from_jdict(jdict.get('opc-version'))
        pmx_model_conf = PMxModelConf.construct_from_jdict(jdict.get('pmx-model-conf'))
        pressure_conf = PressureConf.construct_from_jdict(jdict.get('pressure-conf'))
        psu_conf = PSUConf.construct_from_jdict(jdict.get('psu-conf'))
        psu_version = PSUVersion.construct_from_jdict(jdict.get('psu-version'))
        pt1000_calib = Pt1000Calib.construct_from_jdict(jdict.get('pt1000-calib'))
        scd30_baseline = SCD30Baseline.construct_from_jdict(jdict.get('scd30-baseline'))
        scd30_conf = SCD30Conf.construct_from_jdict(jdict.get('scd30-conf'))
        schedule = Schedule.construct_from_jdict(jdict.get('schedule'))
        shared_secret = SharedSecret.construct_from_jdict(jdict.get('shared-secret'))
        sht_conf = SHTConf.construct_from_jdict(jdict.get('sht-conf'))
        networks = Networks.construct_from_jdict(jdict.get('networks'))
        modem = Modem.construct_from_jdict(jdict.get('modem'))
        sim = SIM.construct_from_jdict(jdict.get('sim'))
        system_id = SystemID.construct_from_jdict(jdict.get('system-id'))
        timezone_conf = TimezoneConf.construct_from_jdict(jdict.get('timezone-conf'))

        return cls(hostname, packs, afe_baseline, afe_id, aws_api_auth,
                   aws_group_config, aws_project, csv_logger_conf, display_conf,
                   vcal_baseline, gas_baseline, gas_model_conf, gps_conf, greengrass_identity,
                   interface_conf, mpl115a2_calib, mqtt_conf, ndir_conf, opc_conf, opc_version,
                   pmx_model_conf, pressure_conf, psu_conf, psu_version, pt1000_calib,
                   scd30_baseline, scd30_conf, schedule, shared_secret, sht_conf, networks,
                   modem, sim, system_id, timezone_conf)


    @classmethod
    def load(cls, manager, psu=None):
        hostname = socket.gethostname()
        packs = PackageVersions.construct_from_installation(manager.scs_path(), manager)

        afe_baseline = AFEBaseline.load(manager)
        afe_id = AFEId.load(manager)
        aws_api_auth = APIAuth.load(manager)
        aws_group_config = AWSGroupConfiguration.load(manager)
        aws_project = Project.load(manager)
        csv_logger_conf = CSVLoggerConf.load(manager)
        display_conf = DisplayConf.load(manager)
        vcal_baseline = VCalBaseline.load(manager)
        gas_baseline = GasBaseline.load(manager)
        gas_model_conf = GasModelConf.load(manager)
        gps_conf = GPSConf.load(manager)
        greengrass_identity = AWSIdentity.load(manager)
        interface_conf = InterfaceConf.load(manager)
        mpl115a2_calib = MPL115A2Calib.load(manager)
        mqtt_conf = MQTTConf.load(manager)
        ndir_conf = NDIRConf.load(manager)
        opc_conf = OPCConf.load(manager)
        opc_version = OPCVersion.load(manager)
        pmx_model_conf = PMxModelConf.load(manager)
        pressure_conf = PressureConf.load(manager)
        psu_conf = PSUConf.load(manager)
        psu_version = None if psu is None else psu.version()
        pt1000_calib = Pt1000Calib.load(manager)
        scd30_baseline = SCD30Baseline.load(manager)
        scd30_conf = SCD30Conf.load(manager)
        schedule = Schedule.load(manager)
        shared_secret = SharedSecret.load(manager)
        sht_conf = SHTConf.load(manager)
        networks = manager.networks()
        modem = manager.modem()
        sim = manager.sim()
        system_id = SystemID.load(manager)
        timezone_conf = TimezoneConf.load(manager)

        return cls(hostname, packs, afe_baseline, afe_id, aws_api_auth,
                   aws_group_config, aws_project, csv_logger_conf, display_conf,
                   vcal_baseline, gas_baseline, gas_model_conf, gps_conf, greengrass_identity,
                   interface_conf, mpl115a2_calib, mqtt_conf, ndir_conf, opc_conf, opc_version,
                   pmx_model_conf, pressure_conf, psu_conf, psu_version, pt1000_calib,
                   scd30_baseline, scd30_conf, schedule, shared_secret, sht_conf, networks,
                   modem, sim, system_id, timezone_conf)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, hostname, packs, afe_baseline, afe_id, aws_api_auth,
                 aws_group_config, aws_project, csv_logger_conf, display_conf,
                 vcal_baseline, gas_baseline, gas_model_conf, gps_conf, greengrass_identity,
                 interface_conf, mpl115a2_calib, mqtt_conf, ndir_conf, opc_conf, opc_version,
                 pmx_model_conf, pressure_conf, psu_conf, psu_version, pt1000_calib,
                 scd30_baseline, scd30_conf, schedule, shared_secret, sht_conf, networks,
                 modem, sim, system_id, timezone_conf):
        """
        Constructor
        """

        self.__hostname = hostname                                  # string
        self.__packs = packs                                        # PackageVersions

        self.__afe_baseline = afe_baseline                          # AFEBaseline
        self.__afe_id = afe_id                                      # AFEId
        self.__aws_api_auth = aws_api_auth                          # APIAuth
        self.__aws_group_config = aws_group_config                  # AWSGroupConfiguration
        self.__aws_project = aws_project                            # Project
        self.__csv_logger_conf = csv_logger_conf                    # CSVLoggerConf
        self.__display_conf = display_conf                          # DisplayConf
        self.__vcal_baseline = vcal_baseline                        # VCalBaseline
        self.__gas_baseline = gas_baseline                          # GasBaseline
        self.__gas_model_conf = gas_model_conf                      # GasModelConf
        self.__gps_conf = gps_conf                                  # GPSConf
        self.__greengrass_identity = greengrass_identity            # AWSIdentity
        self.__interface_conf = interface_conf                      # InterfaceConf
        self.__mpl115a2_calib = mpl115a2_calib                      # MPL115A2Calib
        self.__mqtt_conf = mqtt_conf                                # MQTTConf
        self.__ndir_conf = ndir_conf                                # NDIRConf
        self.__opc_conf = opc_conf                                  # OPCConf
        self.__opc_version = opc_version                            # OPCVersion
        self.__pmx_model_conf = pmx_model_conf                      # PMxModelConf
        self.__pressure_conf = pressure_conf                        # PressureConf
        self.__psu_conf = psu_conf                                  # PSUConf
        self.__psu_version = psu_version                            # PSUVersion
        self.__pt1000_calib = pt1000_calib                          # Pt1000Calib
        self.__scd30_baseline = scd30_baseline                      # SCD30Baseline
        self.__scd30_conf = scd30_conf                              # SCD30Conf
        self.__schedule = schedule                                  # Schedule
        self.__shared_secret = shared_secret                        # SharedSecret
        self.__sht_conf = sht_conf                                  # SHTConf
        self.__networks = networks                                  # Networks
        self.__modem = modem                                        # Modem
        self.__sim = sim                                            # SIM
        self.__system_id = system_id                                # SystemID
        self.__timezone_conf = timezone_conf                        # TimezoneConf


    def __eq__(self, other):
        try:
            return self.hostname == other.hostname and self.packs == other.packs and \
                   self.afe_baseline == other.afe_baseline and self.afe_id == other.afe_id and \
                   self.aws_api_auth == other.aws_api_auth and self.aws_group_config == other.aws_group_config and \
                   self.aws_project == other.aws_project and self.csv_logger_conf == other.csv_logger_conf and \
                   self.display_conf == other.display_conf and self.vcal_baseline == other.vcal_baseline and \
                   self.gas_baseline == other.gas_baseline and self.gas_model_conf == other.gas_model_conf and \
                   self.gps_conf == other.gps_conf and self.greengrass_identity == other.greengrass_identity and \
                   self.interface_conf == other.interface_conf and self.mpl115a2_calib == other.mpl115a2_calib and \
                   self.mqtt_conf == other.mqtt_conf and self.ndir_conf == other.ndir_conf and \
                   self.opc_conf == other.opc_conf and self.pmx_model_conf == other.pmx_model_conf and \
                   self.pmx_model_conf == other.pmx_model_conf and self.pressure_conf == other.pressure_conf and \
                   self.psu_conf == other.psu_conf and self.psu_version == other.psu_version and \
                   self.pt1000_calib == other.pt1000_calib and self.scd30_baseline == other.scd30_baseline and \
                   self.scd30_conf == other.scd30_conf and self.schedule == other.schedule and \
                   self.shared_secret == other.shared_secret and self.sht_conf == other.sht_conf and \
                   self.networks == other.networks and self.modem == other.modem and \
                   self.sim == other.sim and self.system_id == other.system_id and \
                   self.timezone_conf == other.timezone_conf

        except (TypeError, AttributeError):
            return False


    def diff(self, other):
        diff = Configuration(None, None, None, None, None,
                             None, None, None, None, None,
                             None, None, None, None, None,
                             None, None, None, None, None,
                             None, None, None, None, None,
                             None, None, None, None, None,
                             None, None, None, None, None)

        if self.hostname != other.hostname:
            diff.__hostname = self.hostname

        if self.packs != other.packs:
            diff.__packs = self.packs

        if self.afe_id != other.afe_id:
            diff.__afe_id = self.afe_id

        if self.aws_api_auth != other.aws_api_auth:
            diff.__aws_api_auth = self.aws_api_auth

        if self.aws_group_config != other.aws_group_config:
            diff.__aws_group_config = self.aws_group_config

        if self.aws_project != other.aws_project:
            diff.__aws_project = self.aws_project

        if self.csv_logger_conf != other.csv_logger_conf:
            diff.__csv_logger_conf = self.csv_logger_conf

        if self.display_conf != other.display_conf:
            diff.__display_conf = self.display_conf

        if self.vcal_baseline != other.vcal_baseline:
            diff.__vcal_baseline = self.vcal_baseline

        if self.gas_baseline != other.gas_baseline:
            diff.__gas_baseline = self.gas_baseline

        if self.gas_model_conf != other.gas_model_conf:
            diff.__gas_model_conf = self.gas_model_conf

        if self.gps_conf != other.gps_conf:
            diff.__gps_conf = self.gps_conf

        if self.greengrass_identity != other.greengrass_identity:
            diff.__greengrass_identity = self.greengrass_identity

        if self.interface_conf != other.interface_conf:
            diff.__interface_conf = self.interface_conf

        if self.mpl115a2_calib != other.mpl115a2_calib:
            diff.__mpl115a2_calib = self.mpl115a2_calib

        if self.mqtt_conf != other.mqtt_conf:
            diff.__mqtt_conf = self.mqtt_conf

        if self.ndir_conf != other.ndir_conf:
            diff.__ndir_conf = self.ndir_conf

        if self.opc_conf != other.opc_conf:
            diff.__opc_conf = self.opc_conf

        if self.opc_version != other.opc_version:
            diff.__opc_version = self.opc_version

        if self.pmx_model_conf != other.pmx_model_conf:
            diff.__pmx_model_conf = self.pmx_model_conf

        if self.pressure_conf != other.pressure_conf:
            diff.__pressure_conf = self.pressure_conf

        if self.psu_conf != other.psu_conf:
            diff.__psu_conf = self.psu_conf

        if self.psu_version != other.psu_version:
            diff.__psu_version = self.psu_version

        if self.pt1000_calib != other.pt1000_calib:
            diff.__pt1000_calib = self.pt1000_calib

        if self.scd30_baseline != other.scd30_baseline:
            diff.__scd30_baseline = self.scd30_baseline

        if self.scd30_conf != other.scd30_conf:
            diff.__scd30_conf = self.scd30_conf

        if self.schedule != other.schedule:
            diff.__schedule = self.schedule

        if self.shared_secret != other.shared_secret:
            diff.__shared_secret = self.shared_secret

        if self.sht_conf != other.sht_conf:
            diff.__sht_conf = self.sht_conf

        if self.networks != other.networks:
            diff.__networks = self.networks

        if self.modem != other.modem:
            diff.__modem = self.modem

        if self.sim != other.sim:
            diff.__sim = self.sim

        if self.system_id != other.system_id:
            diff.__system_id = self.system_id

        if self.timezone_conf != other.timezone_conf:
            diff.__timezone_conf = self.timezone_conf

        return diff


    # ----------------------------------------------------------------------------------------------------------------

    def save(self, manager):
        if self.hostname:
            raise ValueError('hostname may not be set')

        if self.packs:
            raise ValueError('packs may not be set')

        if self.afe_baseline:
            self.afe_baseline.save(manager)

        if self.afe_id:
            raise ValueError('afe_id may not be set')

        if self.aws_api_auth:
            self.aws_api_auth.save(manager)

        if self.aws_group_config:
            self.aws_group_config.save(manager)

        if self.aws_project:
            self.aws_project.save(manager)

        if self.csv_logger_conf:
            self.csv_logger_conf.save(manager)

        if self.display_conf:
            self.display_conf.save(manager)

        if self.vcal_baseline:
            self.vcal_baseline.save(manager)

        if self.gas_baseline:
            self.gas_baseline.save(manager)

        if self.gas_model_conf:
            self.gas_model_conf.save(manager)

        if self.gps_conf:
            self.gps_conf.save(manager)

        if self.greengrass_identity:
            self.greengrass_identity.save(manager)

        if self.interface_conf:
            self.interface_conf.save(manager)

        if self.mpl115a2_calib:
            self.mpl115a2_calib.save(manager)

        if self.mqtt_conf:
            self.mqtt_conf.save(manager)

        if self.ndir_conf:
            self.ndir_conf.save(manager)

        if self.opc_conf:
            self.opc_conf.save(manager)

        if self.opc_version:
            raise ValueError('opc_version may not be set')

        if self.pmx_model_conf:
            self.pmx_model_conf.save(manager)

        if self.pressure_conf:
            self.pressure_conf.save(manager)

        if self.psu_conf:
            self.psu_conf.save(manager)

        if self.psu_version:
            raise ValueError('psu_version may not be set')

        if self.pt1000_calib:
            self.pt1000_calib.save(manager)

        if self.scd30_baseline:
            self.scd30_baseline.save(manager)

        if self.scd30_conf:
            self.scd30_conf.save(manager)

        if self.schedule:
            self.schedule.save(manager)

        if self.shared_secret:
            self.shared_secret.save(manager)

        if self.sht_conf:
            self.sht_conf.save(manager)

        if self.networks:
            raise ValueError('networks may not be set')

        if self.modem:
            raise ValueError('modem may not be set')

        if self.sim:
            raise ValueError('sim may not be set')

        if self.system_id:
            self.system_id.save(manager)

        if self.timezone_conf:
            self.timezone_conf.save(manager)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['hostname'] = self.hostname
        jdict['packs'] = self.packs

        jdict['afe-baseline'] = self.afe_baseline
        jdict['afe-id'] = self.afe_id
        jdict['aws-api-auth'] = self.aws_api_auth
        jdict['aws-group-config'] = self.aws_group_config
        jdict['aws-project'] = self.aws_project
        jdict['csv-logger-conf'] = self.csv_logger_conf
        jdict['display-conf'] = self.display_conf
        jdict['vcal-baseline'] = self.vcal_baseline
        jdict['gas-baseline'] = self.gas_baseline
        jdict['gas-model-conf'] = self.gas_model_conf
        jdict['gps-conf'] = self.gps_conf
        jdict['interface-conf'] = self.interface_conf
        jdict['greengrass-identity'] = self.greengrass_identity
        jdict['mpl115a2-calib'] = self.mpl115a2_calib
        jdict['mqtt-conf'] = self.mqtt_conf
        jdict['ndir-conf'] = self.ndir_conf
        jdict['opc-conf'] = self.opc_conf
        jdict['opc-version'] = self.opc_version
        jdict['pmx-model-conf'] = self.pmx_model_conf
        jdict['pressure-conf'] = self.pressure_conf
        jdict['psu-conf'] = self.psu_conf
        jdict['psu-version'] = self.psu_version
        jdict['pt1000-calib'] = self.pt1000_calib
        jdict['scd30-baseline'] = self.scd30_baseline
        jdict['scd30-conf'] = self.scd30_conf
        jdict['schedule'] = self.schedule
        jdict['shared-secret'] = self.shared_secret
        jdict['sht-conf'] = self.sht_conf
        jdict['sht-conf'] = self.sht_conf
        jdict['networks'] = self.networks
        jdict['modem'] = self.modem
        jdict['sim'] = self.sim
        jdict['system-id'] = self.system_id
        jdict['timezone-conf'] = self.timezone_conf

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def hostname(self):
        return self.__hostname


    @property
    def packs(self):
        return self.__packs


    @property
    def afe_baseline(self):
        return self.__afe_baseline


    @property
    def afe_id(self):
        return self.__afe_id


    @property
    def aws_api_auth(self):
        return self.__aws_api_auth


    @property
    def aws_group_config(self):
        return self.__aws_group_config


    @property
    def aws_project(self):
        return self.__aws_project


    @property
    def csv_logger_conf(self):
        return self.__csv_logger_conf


    @property
    def display_conf(self):
        return self.__display_conf


    @property
    def vcal_baseline(self):
        return self.__vcal_baseline


    @property
    def gas_baseline(self):
        return self.__gas_baseline


    @property
    def gas_model_conf(self):
        return self.__gas_model_conf


    @property
    def gps_conf(self):
        return self.__gps_conf


    @property
    def greengrass_identity(self):
        return self.__greengrass_identity


    @property
    def interface_conf(self):
        return self.__interface_conf


    @property
    def mpl115a2_calib(self):
        return self.__mpl115a2_calib


    @property
    def mqtt_conf(self):
        return self.__mqtt_conf


    @property
    def ndir_conf(self):
        return self.__ndir_conf


    @property
    def opc_conf(self):
        return self.__opc_conf


    @property
    def opc_version(self):
        return self.__opc_version


    @property
    def pmx_model_conf(self):
        return self.__pmx_model_conf


    @property
    def pressure_conf(self):
        return self.__pressure_conf


    @property
    def psu_conf(self):
        return self.__psu_conf


    @property
    def psu_version(self):
        return self.__psu_version


    @property
    def pt1000_calib(self):
        return self.__pt1000_calib


    @property
    def scd30_baseline(self):
        return self.__scd30_baseline


    @property
    def scd30_conf(self):
        return self.__scd30_conf


    @property
    def schedule(self):
        return self.__schedule


    @property
    def shared_secret(self):
        return self.__shared_secret


    @property
    def sht_conf(self):
        return self.__sht_conf


    @property
    def networks(self):
        return self.__networks


    @property
    def modem(self):
        return self.__modem


    @property
    def sim(self):
        return self.__sim


    @property
    def system_id(self):
        return self.__system_id


    @property
    def timezone_conf(self):
        return self.__timezone_conf


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Configuration:{hostname:%s, packs:%s, afe_baseline:%s, afe_id:%s, aws_api_auth:%s, " \
               "aws_group_config:%s, aws_project:%s, csv_logger_conf:%s, display_conf:%s, " \
               "vcal_baseline:%s, gas_baseline:%s, gas_model_conf:%s, gps_conf:%s, greengrass_identity:%s, " \
               "interface_conf:%s, mpl115a2_calib:%s, mqtt_conf:%s, ndir_conf:%s, opc_conf:%s, " \
               "opc_version:%s, pmx_model_conf:%s, pressure_conf:%s, psu_conf:%s, psu_version:%s, " \
               "pt1000_calib:%s,  scd30_baseline:%s, scd30_conf:%s, schedule:%s, shared_secret:%s, " \
               "sht_conf:%s, networks:%s,  modem:%s, sim:%s, system_id:%s, timezone_conf:%s}" % \
               (self.hostname, self.packs, self.afe_baseline, self.afe_id, self.aws_api_auth,
                self.aws_group_config, self.aws_project, self.csv_logger_conf, self.display_conf,
                self.vcal_baseline, self.gas_baseline, self.gas_model_conf, self.gps_conf, self.greengrass_identity,
                self.interface_conf, self.mpl115a2_calib, self.mqtt_conf, self.ndir_conf, self.opc_conf,
                self.opc_version, self.pmx_model_conf, self.pressure_conf, self.psu_conf, self.psu_version,
                self.pt1000_calib, self.scd30_baseline, self.scd30_conf, self.schedule, self.shared_secret,
                self.sht_conf, self.networks, self.modem, self.sim, self.system_id, self.timezone_conf)



"""
Created on 20 Jan 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://realpython.com/python-logging/
"""

import logging
import sys


# --------------------------------------------------------------------------------------------------------------------

# noinspection PyPep8Naming
class Logging(object):
    """
    classdocs
    """

    __NAME = None

    __MULTI_FORMAT = '%(name)s: %(message)s'

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def config(cls, name, verbose=False, level=logging.ERROR, stream=sys.stderr):
        cls.__NAME = name

        level = logging.INFO if verbose else level

        logging.basicConfig(format=cls.__MULTI_FORMAT, level=level, stream=stream)


    @classmethod
    def getLogger(cls, name=None):
        logger_name = cls.__NAME if cls.__NAME else name

        return logging.getLogger(logger_name)



"""
Created on 4 Oct 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib
"""

import socket

from abc import ABC, abstractmethod
from collections import OrderedDict
from subprocess import Popen, DEVNULL, TimeoutExpired

from scs_core.sys.ipv4_address import IPv4Address


# --------------------------------------------------------------------------------------------------------------------

class Node(ABC):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------
    # network identity...

    @classmethod
    @abstractmethod
    def name(cls):
        pass


    @staticmethod
    def ipv4_address():
        s = socket.socket(family=socket.AF_INET, type=socket.SOCK_DGRAM)

        try:
            s.connect(('192.168.0.1', 1))           # host does not need to be reachable
            dot_decimal = s.getsockname()[0]

        except OSError:
            dot_decimal = '127.0.0.1'

        finally:
            s.close()

        return IPv4Address.construct(dot_decimal)


    @classmethod
    @abstractmethod
    def server_ipv4_address(cls):
        pass


    # ----------------------------------------------------------------------------------------------------------------
    # scanner...

    @classmethod
    def scan_accessible_subnets(cls, start=1, end=254, timeout=10.0):
        for dot_decimal in cls.scan_subnet(cls.server_ipv4_address(), start=start, end=end, timeout=timeout):
            yield dot_decimal

        for dot_decimal in cls.scan_subnet(cls.ipv4_address(), start=start, end=end, timeout=timeout):
            yield dot_decimal


    @classmethod
    def scan_subnet(cls, ipv4_address, start=1, end=254, timeout=10.0):
        if ipv4_address is None:
            return

        pings = OrderedDict()

        # start...
        for addr in ipv4_address.lso_range(start, end):
            dot_decimal = addr.dot_decimal()
            pings[dot_decimal] = Popen(['ping', '-n', '-q', '-c', '1', '-t', str(timeout), dot_decimal],
                                       stdout=DEVNULL, stderr=DEVNULL)
        # report...
        for dot_decimal in pings:
            try:
                pings[dot_decimal].wait(timeout=timeout * 2.0)
            except (TimeoutExpired, TimeoutError):
                continue

            if pings[dot_decimal].returncode == 0:
                yield dot_decimal


    @staticmethod
    def ping(host, timeout=1.0):
        p = Popen(['ping', '-q', '-c', '1', '-t', str(timeout), host],
                  stdout=DEVNULL, stderr=DEVNULL)
        p.wait()

        return p.returncode == 0


    # ----------------------------------------------------------------------------------------------------------------
    # software update...

    @abstractmethod
    def software_update_report(self):
        pass


    # ----------------------------------------------------------------------------------------------------------------
    # filesystem paths...

    @abstractmethod
    def home_path(self):
        pass


    @abstractmethod
    def scs_path(self):
        pass


# --------------------------------------------------------------------------------------------------------------------

class IoTNode(Node):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------
    # identity...

    @classmethod
    def numeric_component_of_name(cls):
        hostname = socket.gethostname()
        pieces = hostname.split('-')

        if len(pieces) != 3:
            raise ValueError(hostname)

        try:
            numeric_component = int(pieces[2])
        except ValueError:
            raise ValueError(hostname)

        return numeric_component


    # ----------------------------------------------------------------------------------------------------------------
    # status...

    @abstractmethod
    def status(self):
        pass


    # ----------------------------------------------------------------------------------------------------------------
    # networks and modem...

    @classmethod
    @abstractmethod
    def networks(cls):
        pass


    @classmethod
    @abstractmethod
    def modem(cls):
        pass


    @classmethod
    @abstractmethod
    def modem_conn(cls):
        pass


    @classmethod
    @abstractmethod
    def sim(cls):
        pass


    # ----------------------------------------------------------------------------------------------------------------
    # SPI...

    @abstractmethod
    def ndir_spi_bus(self):
        pass


    @abstractmethod
    def ndir_spi_device(self):
        pass


    @abstractmethod
    def opc_spi_bus(self):
        pass


    @abstractmethod
    def opc_spi_device(self):
        pass


    # ----------------------------------------------------------------------------------------------------------------
    # time...

    @abstractmethod
    def time_is_synchronized(self):
        pass


    @abstractmethod
    def uptime(self, now=None):
        pass


    # ----------------------------------------------------------------------------------------------------------------
    # tmp directories...

    @abstractmethod
    def lock_dir(self):
        pass


    @abstractmethod
    def tmp_dir(self):
        pass


    # ----------------------------------------------------------------------------------------------------------------
    # filesystem paths...

    @abstractmethod
    def command_path(self):
        pass


    @abstractmethod
    def eep_image(self):
        pass



"""
Created on 6 Nov 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

DEPRECATED: this version of MessageManager for AWS EC2 server with single-part response.

https://xy1eszuu23.execute-api.us-west-2.amazonaws.com/staging/topicMessages?
topic=south-coast-science-dev/production-test/loc/1/gases&startTime=2018-03-31T10:45:50Z&endTime=2018-03-31T10:46:50Z
"""

from scs_core.aws.client.rest_client import RESTClient
from scs_core.aws.data.message import Message

from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class MessageManager(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, api_key, _=None):
        """
        Constructor
        """
        self.__rest_client = RESTClient(api_key)


    # ----------------------------------------------------------------------------------------------------------------

    def find_for_topic(self, topic, start_date, end_date):
        request_path = '/' + '/'.join((topic, start_date.utc().as_iso8601(), end_date.utc().as_iso8601()))

        # request...
        self.__rest_client.connect()

        try:
            jdict = self.__rest_client.get(request_path)

            # messages...
            collection = MessageResponse.construct_from_jdict(jdict)

            messages = [] if collection is None else collection.items

        finally:
            self.__rest_client.close()

        return messages


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MessageManager:{rest_client:%s}" % self.__rest_client


# --------------------------------------------------------------------------------------------------------------------

class MessageResponse(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        items = [Message.construct_from_jdict(msg_jdict) for msg_jdict in jdict.get('Items')]

        count = jdict.get('Count')
        scanned_count = jdict.get('ScannedCount')

        return MessageResponse(items, count, scanned_count)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, items, count, scanned_count):
        """
        Constructor
        """
        self.__items = items                        # list of Message
        self.__count = count                        # int
        self.__scanned_count = scanned_count        # int


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def items(self):
        return self.__items


    @property
    def count(self):
        return self.__count


    @property
    def scanned_count(self):
        return self.__scanned_count


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MessageResponse:{items:%s, count:%s, scanned_count:%s}" % \
               (Str.collection(self.items), self.count, self.scanned_count)



"""
Created on 04 Dec 2020

@author: Jade Page (jade.page@southcoastscience.com)

Case insensitive starts with solution:
https://stackoverflow.com/questions/13578916/case-insensitive-string-startswith-in-python
"""


# --------------------------------------------------------------------------------------------------------------------

class SagemakerModelManager(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, client):
        """
        Constructor
        """
        self.__client = client


    # ----------------------------------------------------------------------------------------------------------------

    def get_models(self, prefix):
        model_list = self.list_model_names_prefix(prefix, None)
        models = []

        for item in model_list:
            res = self.__client.describe_model(
                ModelName=item
            )
            models.append(res)

        return models


    def delete_models_with_prefix(self, prefix):
        deletion_list = self.list_model_names_prefix(prefix, None)

        for item in deletion_list:
            self.__client.delete_model(
                ModelName=item
            )

        return len(deletion_list)


    def list_model_names_prefix(self, prefix, time_order):
        result = []
        names = self.list_model_names_filter(prefix, time_order)

        for name in names:
            if name.startswith(prefix):
                result.append(name)

        return result


    def list_model_names_filter(self, string_filter, time_order):
        names = []
        next_token = None

        while True:
            res, next_token = self.retrieve_filtered_models(string_filter, time_order, next_token)
            models = res.get("Models")

            for item in models:
                name = item.get("ModelName")
                names.append(name)

            if not next_token:
                return names


    def list_model_names(self, time_order):
        names = []
        next_token = None

        while True:
            res, next_token = self.retrieve_models(time_order, next_token)
            models = res.get("Models")

            for item in models:
                name = item.get("ModelName")
                names.append(name)

            if not next_token:
                return names


    def retrieve_models(self, time_order, next_token=None):
        next_token2 = None

        if next_token:
            response = self.__client.list_models(
                SortBy='CreationTime' if time_order else 'Name',
                SortOrder='Descending' if time_order else 'Ascending',
                MaxResults=100,
                NextToken=next_token
            )
        else:
            response = self.__client.list_models(
                SortBy='CreationTime' if time_order else 'Name',
                SortOrder='Descending' if time_order else 'Ascending',
                MaxResults=100
            )

        if "NextToken" in response:
            next_token2 = response.get("NextToken")

        return response, next_token2


    def retrieve_filtered_models(self, filter_string, time_order, next_token=None):
        next_token2 = None

        if next_token:
            response = self.__client.list_models(
                SortBy='CreationTime' if time_order else 'Name',
                SortOrder='Descending' if time_order else 'Ascending',
                MaxResults=100,
                NameContains=filter_string,
                NextToken=next_token
            )
        else:
            response = self.__client.list_models(
                SortBy='CreationTime' if time_order else 'Name',
                SortOrder='Descending' if time_order else 'Ascending',
                MaxResults=100,
                NameContains=filter_string
            )

        if "NextToken" in response:
            next_token2 = response.get("NextToken")

        return response, next_token2


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "SagemakerModelManager:{client:%s}" % self.__client




"""
Created on 1 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

source repo: scs_analysis
"""

import sys

from decimal import InvalidOperation

from scs_core.data.categorical_regression import CategoricalRegression
from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.linear_regression import LinearRegression
from scs_core.data.path_dict import PathDict
from scs_core.data.precision import Precision
from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class Aggregate(object):
    """
    classdocs
    """

    RULE_RATIO = 0.75                                   # ratio of reported points to expected points - 75%

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, min_max, iso_path, nodes):
        """
        Constructor
        """
        self.__min_max = min_max                        # bool
        self.__iso_path = iso_path                      # string
        self.__nodes = nodes                            # list of string

        self.__precisions = {}                          # dict of path: Precision
        self.__regressions = {}                         # dict of path: LinearRegression

        self.__initialised = False                      # bool
        self.__block_sample_count = 0                   # int


    # ----------------------------------------------------------------------------------------------------------------

    def has_value(self):
        for regression in self.__regressions.values():
            if regression.has_midpoint():
                return True

        return False


    def append(self, datetime: LocalizedDatetime, sample: PathDict):
        # initialise...
        if not self.__initialised:
            nodes = self.__nodes if self.__nodes else sample.paths()        # use all paths if nodes are not specified

            for node in nodes:
                try:
                    paths = sample.paths(node)
                except KeyError:
                    continue

                except IndexError as ex:
                    paths = None
                    print("sample_aggregate: %s: IndexError: %s" % (node, ex), file=sys.stderr)
                    sys.stderr.flush()
                    exit(1)

                for path in paths:
                    if path == 'rec':
                        continue

                    self.__precisions[path] = Precision()
                    self.__regressions[path] = CategoricalRegression() if isinstance(sample.node(path), str) else \
                        LinearRegression()

            self.__initialised = True

        # values...
        for path in self.__precisions.keys():
            try:
                value = sample.node(path)
            except KeyError:
                continue

            if value is None:
                continue

            try:
                self.__precisions[path].widen(value)
                self.__regressions[path].append(datetime, value)
            except InvalidOperation:
                continue

        self.__block_sample_count += 1


    def reset(self):
        self.__block_sample_count = 0

        for path in self.__regressions.keys():
            self.__regressions[path].reset()


    def report(self, localised_datetime):
        report = PathDict()

        # rec...
        report.append(self.__iso_path, localised_datetime.as_iso8601())

        # values...
        for path, precision in self.__precisions.items():
            regression = self.__regressions[path]

            if self.__regressions[path].has_midpoint():
                if self.__min_max:
                    report.append(path + '.min', regression.min(precision.digits))
                    report.append(path + '.mid', regression.mid(precision.digits))
                    report.append(path + '.max', regression.max(precision.digits))

                else:
                    report.append(path, regression.mid(precision.digits))

        return report


    # ----------------------------------------------------------------------------------------------------------------

    def complies_with_rule(self, sampling_interval, aggregate_interval):
        expected_count = aggregate_interval.total_seconds() / sampling_interval.total_seconds()
        ratio = self.block_sample_count / expected_count

        return ratio >= self.RULE_RATIO


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def block_sample_count(self):
        return self.__block_sample_count


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Aggregate:{min_max:%s, iso_path:%s block_sample_count:%s, regressions:%s}" % \
               (self.__min_max, self.__iso_path, self.block_sample_count, Str.collection(self.__regressions))



"""
Created on 30 Apr 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
    {
      "topic": "/orgs/south-coast-science-dev/user/device/alpha-pi-eng-000100/status",
      "name": "Device status",
      "description": "lat (deg), lng (deg) GPS qual, DFE temp (Centigrade), host temp (Centigrade), errors",
      "public": true,
      "topic-info": {
        "format": "application/json"
      },
      "unit": null,
      "stats": {
        "period": 0,
        "last-reading": "55 days ago",
        "contributors": [
          {
            "name": "South Coast Science - Dev",
            "id": "southcoastscience-dev",
            "gravatar-hash": "07f512e9fe64863039df0c0f1834cc25"
          }
        ],
        "last-location": {
          "lat": 50.82313,
          "lon": -0.122922
        }
      },
      "owner-gravatar-hash": "d92c0a697bf3658745447e1a9467b506"
    }
"""

from scs_core.osio.data.abstract_topic import AbstractTopic
from scs_core.osio.data.topic_info import TopicInfo
from scs_core.osio.data.topic_stats import TopicStats


# --------------------------------------------------------------------------------------------------------------------

class UserTopic(AbstractTopic):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        # AbstractTopic...
        path = jdict.get('topic')
        name = jdict.get('name')
        description = jdict.get('description')

        is_public = jdict.get('public')

        info = TopicInfo.construct_from_jdict(jdict.get('topic-info'))

        # UserTopic...
        unit = jdict.get('unit')
        stats = TopicStats.construct_from_jdict(jdict.get('stats'))

        owner_gravatar_hash = jdict.get('owner-gravatar-hash')

        return UserTopic(path, name, description, is_public, info, unit, stats, owner_gravatar_hash)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, path, name, description, is_public, info,
                 unit, stats, owner_gravatar_hash):
        """
        Constructor
        """

        # AbstractTopic...
        AbstractTopic.__init__(self, path, name, description, is_public, info)

        # UserTopic...
        self.__unit = unit                                      # string
        self.__stats = stats                                    # TopicStats
        self.__owner_gravatar_hash = owner_gravatar_hash        # string


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = AbstractTopic.as_json(self)

        jdict['unit'] = self.unit
        jdict['stats'] = self.stats
        jdict['owner-gravatar-hash'] = self.owner_gravatar_hash

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def unit(self):
        return self.__unit


    @property
    def stats(self):
        return self.__stats


    @property
    def owner_gravatar_hash(self):
        return self.__owner_gravatar_hash


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "UserTopic:{path:%s, name:%s, description:%s, is_public:%s, info:%s, " \
               "unit:%s, stats:%s, owner_gravatar_hash:%s}" % \
               (self.path, self.name, self.description, self.is_public, self.info,
                self.unit, self.stats, self.owner_gravatar_hash)



"""
Created on 22 Mar 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

schemas:
{"id": 33, "name": "south-coast-science-climate",
"description": "South Coast Science Air Quality Sensor; climate"}

{"id": 29, "name": "south-coast-science-particulates", "description":
"South Coast Science Air Quality Sensor; particulates"}


{"id": 28, "name": "south-coast-science-gasses-Ox-NO2-NO-CO",
"description": "South Coast Science Air Quality Sensor; concentration of gasses"}

{"id": 34, "name": "south-coast-science-gasses-NO2-CO-SO2-H2S",
"description": "South Coast Science Air Quality Sensor; concentration of gasses"}

{"id": 35, "name": "south-coast-science-gasses-NO-Ox-CO-SO2",
"description": "South Coast Science Air Quality Sensor; concentration of gasses"}

{"id": 39, "name": "south-coast-science-gasses-NO2-Ox-CO-SO2",
"description": "South Coast Science Air Quality Sensor; concentration of gasses"}
"""

from collections import OrderedDict

from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class ProjectTopic(JSONable):
    """
    classdocs
    """

    CLIMATE = None
    PARTICULATES = None
    STATUS = None
    CONTROL = None
    EMPTY_GASES_TOPIC = None

    __GASES_TOPICS = {}

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def init(cls):
        cls.CLIMATE = ProjectTopic(33, 'Climate',
                                   'temperature (Centigrade), relative humidity (%)', ('temperature', 'humidity'))

        cls.PARTICULATES = ProjectTopic(29, 'Particulate densities',
                                        'PM1 (ug/m3), PM2.5 (ug/m3), PM10 (ug/m3), bin counts, mtf1, mtf3, mtf5 mtf7',
                                        ('PM1', 'PM2.5', 'PM10'))

        cls.STATUS = ProjectTopic(None, 'Device status',
                                  'lat (deg), lng (deg) GPS qual, DFE temp (Centigrade), host temp (Centigrade), ' 
                                  'errors', ())

        cls.CONTROL = ProjectTopic(None, 'Device control',
                                   'this topic is subscribed to by the device for control purposes', ())

        cls.__GASES_TOPICS = {
            ('NO2', 'Ox', 'NO', 'CO'):
                ProjectTopic(28, 'Gas concentrations',
                             'NO2, O3, NO, CO electrochemical we (V), ae (V), wc (V), cnc (ppb), '
                             'Pt1000 temp, internal SHT', ('NO2', 'O3', 'NO', 'CO')),

            ('NO2', 'CO', 'SO2', 'H2S'):
                ProjectTopic(34, 'Gas concentrations',
                             'NO2, CO, SO2, H2S electrochemical we (V), ae (V), wc (V), cnc (ppb), '
                             'Pt1000 temp, internal SHT', ('NO2', 'CO', 'SO2', 'H2S')),

            ('NO', 'Ox', 'CO', 'SO2'):                          # not a standard Alphasense configuration
                ProjectTopic(35, 'Gas concentrations',
                             'NO2, O3, CO, SO2 electrochemical we (V), ae (V), wc (V), cnc (ppb), '
                             'Pt1000 temp, internal SHT', ('NO', 'O3', 'CO', 'SO2')),

            ('NO2', 'Ox', 'CO', 'SO2'):
                ProjectTopic(39, 'Gas concentrations',
                             'NO2, O3, CO, SO2 electrochemical we (V), ae (V), wc (V), cnc (ppb), '
                             'Pt1000 temp, internal SHT', ('NO2', 'O3', 'CO', 'SO2')),

            ('NO2', 'Ox', 'CO', 'VOC'):
                ProjectTopic(None, 'Gas concentrations',
                             'NO2, O3, CO, VOC electrochemical we (V), ae (V), wc (V), cnc (ppb), '
                             'Pt1000 temp, internal SHT', ('NO2', 'O3', 'CO', 'VOC')),

            ('CO', 'SO2', 'H2S', 'VOC'):
                ProjectTopic(None, 'Gas concentrations',
                             'CO, SO2, H2S, VOCs electrochemical we (V), ae (V), wc (V), cnc (ppb), '
                             'Pt1000 temp, internal SHT', ('CO', 'SO2', 'H2S', 'VOC')),

            ('SN1', 'SN2', 'SN3', 'SN4'):
                ProjectTopic(None, 'Test load',
                             'electrochemical we (V), ae (V), '
                             'Pt1000 temp, internal SHT', ('test',))
        }

        cls.EMPTY_GASES_TOPIC = ProjectTopic(None, 'Gas concentrations',
                                             'no information available', ())


    @classmethod
    def get_gases_topic(cls, gas_names):
        if gas_names is None:
            return cls.EMPTY_GASES_TOPIC

        for topic_gas_names, topic in cls.__GASES_TOPICS.items():
            if set(topic_gas_names) == set(gas_names):
                return topic

        return cls.EMPTY_GASES_TOPIC


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, schema_id, name, description, tags):
        """
        Constructor
        """
        self.__schema_id = schema_id            # int
        self.__name = name                      # string
        self.__description = description        # string
        self.__tags = tags                      # sequence of string


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['schema-id'] = self.schema_id
        jdict['name'] = self.name
        jdict['description'] = self.description
        jdict['tags'] = self.tags

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def schema_id(self):
        return self.__schema_id


    @property
    def name(self):
        return self.__name


    @property
    def description(self):
        return self.__description


    @property
    def tags(self):
        return self.__tags


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ProjectTopic:{schema_id:%s, name:%s, tags:%s, description:%s}" % \
               (self.schema_id, self.name, self.tags, self.description)



"""
Created on 24 Dec 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
{"upload": "2018-12-24T13:09:03Z", "rec": "2018-12-24T13:09:01Z", "offset": 2}
"""

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable
from scs_core.data.path_dict import PathDict


# --------------------------------------------------------------------------------------------------------------------

class UploadInterval(JSONable):
    """
    classdocs
    """

    UPLOAD_FIELD =      'upload'
    REC_FIELD =         'payload.rec'

    INCLUDE_MILLIS =    False


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jstr(cls, jstr):
        if not jstr:
            return None

        # document...
        document = PathDict.construct_from_jstr(jstr)

        if not document:
            return None

        # upload...
        upload_node = document.node(cls.UPLOAD_FIELD)
        upload = LocalizedDatetime.construct_from_iso8601(upload_node)

        if upload is None:
            raise ValueError(upload_node)

        # rec...
        rec_node = document.node(cls.REC_FIELD)
        rec = LocalizedDatetime.construct_from_iso8601(rec_node)

        if rec is None:
            raise ValueError(rec_node)

        # offset...
        offset = upload - rec

        return UploadInterval(upload, rec, offset)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, upload, rec, offset):
        """
        Constructor
        """
        self.__upload = upload              # LocalizedDatetime
        self.__rec = rec                    # LocalizedDatetime

        self.__offset = offset              # Timedelta


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['upload'] = self.upload.as_iso8601(include_millis=self.INCLUDE_MILLIS)
        jdict['rec'] = self.rec.as_iso8601(include_millis=self.INCLUDE_MILLIS)

        jdict['offset'] = self.offset

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def upload(self):
        return self.__upload


    @property
    def rec(self):
        return self.__rec


    @property
    def offset(self):
        return self.__offset


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "UploadInterval:{upload:%s, rec:%s, offset:%s}" % (self.upload, self.rec, self.offset)



"""
Created on 23 Dec 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

an ML model group configuration for particulates inference

example JSON:
{"uds-path": "pipes/lambda-model-pmx-s1.uds", "model-interface": "s1"}
"""

from scs_core.model.model_conf import ModelConf
from scs_core.model.pmx.pmx_inference_client import PMxInferenceClient


# --------------------------------------------------------------------------------------------------------------------

class PMxModelConf(ModelConf):
    """
    classdocs
    """

    __FILENAME = "pmx_model_conf.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    __INTERFACES = ('s1', 's2')

    @classmethod
    def interfaces(cls):
        return cls.__INTERFACES


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, uds_path, model_interface, model_compendium_group=None):
        """
        Constructor
        """
        super().__init__(uds_path, model_interface, model_compendium_group=model_compendium_group)


    # ----------------------------------------------------------------------------------------------------------------

    def client(self, host, socket) -> PMxInferenceClient:
        if self.model_interface == 's1' or self.model_interface == 's2':
            from scs_core.model.pmx.s1.s1_pmx_inference_client import S1PMxInferenceClient

            return S1PMxInferenceClient.construct(socket, self.abs_uds_path(host))

        raise ValueError(self.model_interface)



"""
Created on 17 Feb 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document:
{"location-path": "/orgs/southcoastscience-dev/test/loc/1", "device-path": "/orgs/southcoastscience-dev/test/device"}
"""

from collections import OrderedDict

from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class Project(PersistentJSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    __FILENAME =                "osio_project.json"

    @classmethod
    def persistence_location(cls):
        return cls.osio_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def is_valid_channel(cls, channel):
        return channel in ('C', 'G', 'P', 'S', 'X')


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, org_id, group, location_id):
        path_root = '/orgs/' + org_id + '/'

        location_path = path_root + group + '/loc/' + str(location_id)
        device_path = path_root + group + '/device'

        return Project(location_path, device_path)


    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        location_path = jdict.get('location-path')
        device_path = jdict.get('device-path')

        return Project(location_path, device_path)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, location_path, device_path):
        """
        Constructor
        """
        super().__init__()

        self.__location_path = location_path          # string
        self.__device_path = device_path              # string


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['location-path'] = self.__location_path
        jdict['device-path'] = self.__device_path

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def channel_path(self, channel, system_id):
        if channel == 'C':
            return self.climate_topic_path()

        if channel == 'G':
            return self.gases_topic_path()

        if channel == 'P':
            return self.particulates_topic_path()

        if channel == 'S':
            return self.status_topic_path(system_id)

        if channel == 'X':
            return self.control_topic_path(system_id)

        raise ValueError("channel_path: unrecognised channel: %s" % channel)


    # ----------------------------------------------------------------------------------------------------------------

    def climate_topic_path(self):
        return self.__location_path + '/climate'


    def gases_topic_path(self):
        return self.__location_path + '/gases'


    def particulates_topic_path(self):
        return self.__location_path + '/particulates'


    def status_topic_path(self, system_id):
        return '/'.join((self.__device_path, system_id.topic_label(), 'status'))


    def control_topic_path(self, system_id):
        return '/'.join((self.__device_path, system_id.topic_label(), 'control'))


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def location_path(self):
        return self.__location_path


    @property
    def device_path(self):
        return self.__device_path


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Project:{location_path:%s, device_path:%s}" % (self.location_path, self.device_path)
